import pytest

from io import BytesIO
from psi_athena.athena import Athena
from test_utils.mock import MockApi
from typing import Tuple


class TestAthena_UnitTests:

    @pytest.mark.unit
    def test_assume_role(self):
        # Set expected values
        expected_role = 'aws:role:arn'
        expected_region = 'region-code'
        expected_session_name = 'test session name'
        expected_access_key = 'accesskey'
        expected_secret_access_key = 'secret'
        expected_session_token = 'token'

        # Set up mocks
        mock_session, _ = self.get_mocks()
        mock_session.on(
            'assume_role',
            RoleArn=expected_role,
            RoleSessionName=expected_session_name
        ).return_val({
            'Credentials': {
                'AccessKeyId': expected_access_key,
                'SecretAccessKey': expected_secret_access_key,
                'SessionToken': expected_session_token
            }
        })
        mock_session.on(
            'client',
            'athena',
            region_name=expected_region,
            aws_access_key_id=expected_access_key,
            aws_secret_access_key=expected_secret_access_key,
            aws_session_token=expected_session_token
        ).return_val(None)

        # Test assume_role()
        a = Athena(aws_session=mock_session)
        a.assume_role(expected_role, expected_region, expected_session_name)
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_run_athena_query_exception(self):
        # Set expected values
        expected_query = 'sql'
        expected_database = 'test_db'
        expected_table = 'test_table'
        expected_catalog = 'TestCatalog'
        expected_workgroup = 'test-wg'
        expected_query_execution_id = 'expected-id'
        expected_bucket = 'bucket'
        expected_key = 'key'
        expected_exception = Exception('test exception')

        # Set up mocks
        mock_session, mock_athena_client = self.get_mocks()
        mock_athena_client.on(
            'start_query_execution',
            QueryString=expected_query,
            QueryExecutionContext={
                'Database': expected_database,
                'Catalog': expected_catalog
            },
            WorkGroup=expected_workgroup
        ).return_val({'QueryExecutionId': expected_query_execution_id})
        mock_athena_client.on(
            'get_query_execution',
            QueryExecutionId=expected_query_execution_id
        ).return_val({
            'QueryExecution': {
                'Status': {'State': 'SUCCEEDED'},
                'ResultConfiguration': {'OutputLocation': f's3://{expected_bucket}/{expected_key}'}
            }
        })
        mock_s3_client = MockApi()
        mock_s3_client.on(
            'get_object',
            Bucket=expected_bucket,
            Key=expected_key
        ).raise_exception(expected_exception)
        mock_session.on('client','s3').return_val(mock_s3_client)

        # Test run_athena_query() exception
        a = Athena(
            data_catalog=expected_catalog,
            workgroup=expected_workgroup,
            database=expected_database,
            aws_session=mock_session
        )
        with pytest.raises(Exception) as e_info:
            a.run_athena_query(expected_table, expected_query, True)
        assert e_info.value == expected_exception
        mock_athena_client.assert_expectations()
        mock_s3_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_start_query(self):
        # Set expected values
        expected_query = 'sql'
        expected_database = 'test_db'
        expected_catalog = 'TestCatalog'
        expected_workgroup = 'test-wg'
        expected_query_execution_id = 'expected-id'

        # Set up mocks
        mock_session, mock_athena_client = self.get_mocks()
        mock_athena_client.on(
            'start_query_execution',
            QueryString=expected_query,
            QueryExecutionContext={
                'Database': expected_database,
                'Catalog': expected_catalog
            },
            WorkGroup=expected_workgroup
        ).return_val({'QueryExecutionId': expected_query_execution_id})

        # Test start_query()
        a = Athena(
            data_catalog=expected_catalog,
            workgroup=expected_workgroup,
            database=expected_database,
            aws_session=mock_session
        )
        query_execution_id = a.start_query(expected_query)
        assert query_execution_id == expected_query_execution_id
        mock_athena_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_start_query_exception(self):
        # Set expected values
        expected_query = 'sql'
        expected_database = 'test_db'
        expected_catalog = 'TestCatalog'
        expected_workgroup = 'test-wg'
        expected_exception = Exception('test')

        # Set up mocks
        mock_session, mock_athena_client = self.get_mocks()
        mock_athena_client.on(
            'start_query_execution',
            QueryString=expected_query,
            QueryExecutionContext={
                'Database': expected_database,
                'Catalog': expected_catalog
            },
            WorkGroup=expected_workgroup
        ).raise_exception(expected_exception)

        # Test start_query() exception
        a = Athena(
            data_catalog=expected_catalog,
            workgroup=expected_workgroup,
            database=expected_database,
            aws_session=mock_session
        )
        with pytest.raises(Exception) as e_info:
            a.start_query(expected_query)
        assert e_info.value == expected_exception
        mock_athena_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_poll_for_query_results(self):
        # Set expected values
        expected_query_execution_id = 'expected-id'
        expected_results = {'QueryExecution': {'Status': {'State': 'SUCCEEDED'}}}

        # Set up mocks
        mock_session, mock_athena_client = self.get_mocks()
        mock_athena_client.on(
            'get_query_execution',
            QueryExecutionId=expected_query_execution_id
        ).return_val(expected_results)

        # Test poll_for_query_results()
        a = Athena(aws_session=mock_session)
        results = a.poll_for_query_results(expected_query_execution_id)
        assert results == expected_results
        mock_athena_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_poll_for_query_results_exception(self):
        # Set expected values
        expected_query_execution_id = 'expected-id'
        expected_exception = Exception('test')

        # Set up mocks
        mock_session, mock_athena_client = self.get_mocks()
        mock_athena_client.on(
            'get_query_execution',
            QueryExecutionId=expected_query_execution_id
        ).raise_exception(expected_exception)

        # Test poll_for_query_results() exception
        a = Athena(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            a.poll_for_query_results(expected_query_execution_id)
        assert e_info.value == expected_exception
        mock_athena_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_download_query_results_as_dict(self):
        # Set expected values
        expected_bucket = 'bucket'
        expected_key = 'key'
        expected_body = 'test,csv,body\nalpha.val,,4.0\nbeta.val,-10,'.encode('utf-8')

        # Set up mocks
        mock_s3_client = MockApi()
        mock_s3_client.on(
            'get_object',
            Bucket=expected_bucket,
            Key=expected_key
        ).return_val({'Body': BytesIO(expected_body)})
        mock_session, _ = self.get_mocks()
        mock_session.on('client','s3').return_val(mock_s3_client)

        # Test download_query_results_as_dict()
        a = Athena(aws_session=mock_session)
        results = a.download_query_results_as_dict({
            'QueryExecution': {
                'ResultConfiguration': {
                    'OutputLocation': f's3://{expected_bucket}/{expected_key}'
                }
            }
        })
        assert results == [{
            'test': 'alpha.val',
            'csv': None,
            'body': 4.0
        }, {
            'test': 'beta.val',
            'csv': -10,
            'body': None
        }]
        mock_s3_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_download_query_results_as_dict_exception(self):
        # Set expected values
        expected_bucket = 'bucket'
        expected_key = 'key'
        expected_body = 'test,csv,body\nalpha.val,-10,4.0,extra,columns'.encode('utf-8')

        # Set up mocks
        mock_s3_client = MockApi()
        mock_s3_client.on(
            'get_object',
            Bucket=expected_bucket,
            Key=expected_key
        ).return_val({'Body': BytesIO(expected_body)})
        mock_session, _ = self.get_mocks()
        mock_session.on('client','s3').return_val(mock_s3_client)

        # Test download_query_results_as_dict()
        a = Athena(aws_session=mock_session)
        with pytest.raises(AssertionError) as e_info:
            a.download_query_results_as_dict({
                'QueryExecution': {
                    'ResultConfiguration': {
                        'OutputLocation': f's3://{expected_bucket}/{expected_key}'
                    }
                }
            })
        assert str(e_info.value).startswith('Invalid CSV column')
        mock_s3_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_download_query_results(self):
        # Set expected values
        expected_bucket = 'bucket'
        expected_key = 'key'
        expected_body = 'test,csv,body'.encode('utf-8')

        # Set up mocks
        mock_s3_client = MockApi()
        mock_s3_client.on(
            'get_object',
            Bucket=expected_bucket,
            Key=expected_key
        ).return_val({'Body': BytesIO(expected_body)})
        mock_session, _ = self.get_mocks()
        mock_session.on('client','s3').return_val(mock_s3_client)

        # Test download_query_results()
        a = Athena(aws_session=mock_session)
        results = a.download_query_results({
            'QueryExecution': {
                'ResultConfiguration': {
                    'OutputLocation': f's3://{expected_bucket}/{expected_key}'
                }
            }
        })
        assert results == expected_body
        mock_s3_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_download_query_results_exception(self):
        # Set expected values
        expected_bucket = 'bucket'
        expected_key = 'key'
        expected_exception = Exception('test exception')

        # Set up mocks
        mock_s3_client = MockApi()
        mock_s3_client.on(
            'get_object',
            Bucket=expected_bucket,
            Key=expected_key
        ).raise_exception(expected_exception)
        mock_session, _ = self.get_mocks()
        mock_session.on('client','s3').return_val(mock_s3_client)

        # Test download_query_results() exception
        a = Athena(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            a.download_query_results({
                'QueryExecution': {
                    'ResultConfiguration': {
                        'OutputLocation': f's3://{expected_bucket}/{expected_key}'
                    }
                }
            })
        assert e_info.value == expected_exception
        mock_s3_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_reload_partitions_exception(self):
        # Set expected values
        expected_table = 'test_table'
        expected_exception = Exception('test')

        # Set up mocks
        mock_session, mock_athena_client = self.get_mocks()
        mock_athena_client.on(
            'start_query_execution',
            QueryString=f'MSCK REPAIR TABLE `{expected_table}`',
            QueryExecutionContext={
                'Database': 'default',
                'Catalog': 'AwsDataCatalog'
            },
            WorkGroup='primary'
        ).raise_exception(expected_exception)

        # Test reload_partitions() exception
        a = Athena(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            a.reload_partitions(expected_table)
        assert e_info.value == expected_exception
        mock_athena_client.assert_expectations()
        mock_session.assert_expectations()

    def get_mocks(self) -> Tuple[MockApi, MockApi]:
        mock_athena_client = MockApi()
        mock_session = MockApi()
        mock_session.on('client','athena').return_val(mock_athena_client)
        return mock_session, mock_athena_client



import boto3
import csv
import logging
import time

from datetime import datetime
from psi_athena import DEFAULT_LOGGER, log_it
from typing import List, Union

DEFAULT_ATHENA_DATA_CATALOG = 'AwsDataCatalog'
DEFAULT_ATHENA_DATABASE = 'default'
DEFAULT_ATHENA_WORKGROUP = 'primary'
DEFAULT_AWS_REGION = 'us-east-1'
DEFAULT_AWS_SESSION_NAME = 'psi-athena-session'
DEFAULT_MAX_POLL_DURATION_SECONDS = 300
DEFAULT_POLL_WAIT_SECONDS = 5


class Athena:

    def __init__(
        self,
        data_catalog: str = None,
        workgroup: str = None,
        database: str = None,
        max_poll_duration_seconds: int = None,
        poll_wait_seconds: int = None,
        aws_session: boto3.Session = None,
        logger: logging.Logger = None
    ):
        self.data_catalog = data_catalog or DEFAULT_ATHENA_DATA_CATALOG
        self.workgroup = workgroup or DEFAULT_ATHENA_WORKGROUP
        self.database = database or DEFAULT_ATHENA_DATABASE
        self.max_poll_duration_seconds = max_poll_duration_seconds or DEFAULT_MAX_POLL_DURATION_SECONDS
        self.poll_wait_seconds = poll_wait_seconds or DEFAULT_POLL_WAIT_SECONDS
        self.__aws_session = aws_session or boto3.Session()
        self.__athena_client = None
        self.__logger = logger or DEFAULT_LOGGER
        self.assume_role()

    def assume_role(self, role_arn: str = None, region: str = None, session_name: str = None):
        if role_arn:
            region = region or DEFAULT_AWS_REGION
            session_name = session_name or DEFAULT_AWS_SESSION_NAME
            self.__logger.info(f'Assuming role: {role_arn} in region {region}')
            stsresponse = self.__aws_session.assume_role(
                RoleArn=role_arn,
                RoleSessionName=session_name
            )

            self.__logger.debug(f'Initializing Athena client from role: {role_arn}')
            self.__athena_client = self.__aws_session.client(
                'athena',
                region_name=region,
                aws_access_key_id=stsresponse["Credentials"]["AccessKeyId"],
                aws_secret_access_key=stsresponse["Credentials"]["SecretAccessKey"],
                aws_session_token=stsresponse["Credentials"]["SessionToken"]
            )
        else:
            self.__logger.debug('Initializing Athena client from existing AWS session')
            self.__athena_client = self.__aws_session.client('athena')

    @log_it
    def run_athena_query(
        self,
        table: str,
        query: str,
        download_results: bool = True,
        reload_partitions: bool = False
    ) -> Union[dict, list]:
        assert table and isinstance(table, str), f'Invalid table name: {table}'
        assert query and isinstance(query, str), f'Invalid Athena query sql: {query}'
        if reload_partitions:
            self.reload_partitions(table)
        query_execution_id = self.start_query(query)
        query_results = self.poll_for_query_results(query_execution_id)
        if download_results:
            return self.download_query_results_as_dict(query_results)
        return query_results

    def start_query(
        self,
        query: str,
        data_catalog: str = None,
        workgroup: str = None,
        database: str = None
    ) -> str:
        data_catalog = data_catalog or self.data_catalog
        workgroup = workgroup or self.workgroup
        database = database or self.database
        assert query and isinstance(query, str), f'Invalid Athena query: {query}'
        assert data_catalog and isinstance(data_catalog, str), f'Invalid Athena data catalog: {data_catalog}'
        assert workgroup and isinstance(workgroup, str), f'Invalid Athena workgroup: {workgroup}'
        assert database and isinstance(database, str), f'Invalid Athena database: {database}'
        self.__logger.info(f'Starting Athena query: {query}')
        try:
            query_execution_id = self.__athena_client.start_query_execution(
                QueryString=query,
                QueryExecutionContext={
                    'Database': database,
                    'Catalog': data_catalog
                },
                WorkGroup=workgroup
            )['QueryExecutionId']
        except Exception as e:
            self.__logger.error(f'Failed to start Athena query execution: {e}')
            raise e
        self.__logger.info(f'Athena query execution id: {query_execution_id}')
        return query_execution_id

    def poll_for_query_results(self, query_execution_id: str) -> dict:
        assert query_execution_id and isinstance(query_execution_id, str), f'Invalid query_execution_id to poll: {query_execution_id}'
        t0 = datetime.now()
        query_execution_status = 'QUEUED'
        while (datetime.now() - t0).total_seconds() < self.max_poll_duration_seconds:
            try:
                query_results = self.__athena_client.get_query_execution(
                    QueryExecutionId=query_execution_id
                )
                query_execution_status = query_results['QueryExecution']['Status']['State']
            except Exception as e:
                self.__logger.error(f'Failed to retrieve Athena query execution status: {e}')
                raise e

            if query_execution_status in ['SUCCEEDED', 'FAILED', 'CANCELLED']:
                break
            time.sleep(self.poll_wait_seconds)
        assert query_execution_status == 'SUCCEEDED', f'Athena query execution did not succeed, polling exited with status: {query_execution_status}'
        return query_results

    def download_query_results_as_dict(self, query_results: dict) -> List[dict]:
        data = self.download_query_results(query_results)
        try:
            csvstr = data.decode('utf-8')
            parsed = list(csv.DictReader(csvstr.split('\n')))
            for i in range(len(parsed)):
                for k, v in parsed[i].items():
                    assert k and isinstance(k, str), f'Invalid CSV column (likely that extra columns found): {k}'
                    if len(v) == 0:
                        # Parse empty strings as None
                        parsed[i][k] = None
                    elif '.' in v:
                        # Cast numbers with radix point as floats
                        try:
                            float_val = float(v)
                            parsed[i][k] = float_val
                        except ValueError:
                            pass
                    else:
                        # Cast numbers as ints
                        try:
                            int_val = int(v)
                            parsed[i][k] = int_val
                        except ValueError:
                            pass
            return parsed
        except Exception as e:
            self.__logger.error(f'Failed to parse query results into a dict: {e}')
            raise e

    def download_query_results(self, query_results: dict) -> bytes:
        assert query_results and isinstance(query_results, dict), f'Invalid query_results to download results for: {query_results}'
        s3_path = query_results.get('QueryExecution', {}).get('ResultConfiguration', {}).get('OutputLocation')
        try:
            client = self.__aws_session.client('s3')
            self.__logger.debug(f'Fetching object from S3: {s3_path}')
            obj = client.get_object(
                Bucket=s3_path[len('s3://'):].split('/', 1)[0],
                Key=s3_path[len('s3://'):].split('/', 1)[1]
            )
            return obj['Body'].read()
        except Exception as e:
            self.__logger.error(f'Failed to download results with error: {e}')
            raise e

    def reload_partitions(self, table: str):
        assert table and isinstance(table, str), f'Invalid table name: {table}'
        query = f'MSCK REPAIR TABLE `{table}`'
        query_result = self.run_athena_query(
            table,
            query,
            download_results=False,
            reload_partitions=False
        )
        assert query_result['QueryExecution']['Status']['State'] == 'SUCCEEDED'
